<?php

namespace Drupal\plugin_config_form_builder\Element;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\FormElementBase;
use Drupal\Component\Plugin\PluginManagerInterface;

/**
 * Provides a plugin_config element.
 */
#[FormElement('plugin_config')]
class PluginConfig extends FormElementBase {


  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processPlugin'],
      ],
      '#theme_wrappers' => ['container'],
    ];
  }

  /**
   * Expands the element.
   */
  public static function processPlugin(&$element, FormStateInterface $form_state, &$complete_form) {
    /** @var PluginManagerInterface $plugin_manager */
    $plugin_manager = \Drupal::service($element['#plugin_manager']);

    $wrapper = 'plugin-config-' . ($element['#uuid'] ?? $element['#id']) . '-wrapper';
    $element += [
      '#type' => 'fieldset',
      '#collapsible' => FALSE,
      '#prefix' => '<div id="' . $wrapper . '">',
      '#suffix' => '</div>',
    ];
    $element['#tree'] = TRUE;
    if (empty($element['#vertical_tabs']) || $element['#vertical_tabs']) {
      $element['plugin_group'] = [
        '#type' => 'vertical_tabs',
      ];
    }

    $plugins = $form_state->getValue($element['#parents']) ?: $element['#default_value'];
    $plugins = array_filter($plugins, function($p) {return is_array($p) && isset($p['id']);});
    // If we are on an ajax request we assume that a new item has been added.
    $new_plugin_uuid = isset($form_state->getTriggeringElement()['#ajax']) ? array_key_last($plugins) : array_key_first($plugins);
    foreach ($plugins as $uuid => $plugin) {
      $element[$uuid] = [
        '#type' => 'details',
        '#title' => $plugin_manager->getDefinition($plugin['id'])['label'],
        '#group' => implode('][', $element['#array_parents']) . '][plugin_group',
        '#open' => $uuid === $new_plugin_uuid,
      ];
      $element[$uuid]['configuration'] = [
        '#type' => 'single_plugin_config',
        '#title' => t('Configuration'),
        '#plugin_manager' => $element['#plugin_manager'],
        '#default_value' => $plugin,
        '#parents' => array_merge($element['#array_parents'], [$uuid]),
      ];
      $element[$uuid]['id'] = [
        '#type' => 'hidden',
        '#default_value' => $plugin['id'],
      ];
      $element[$uuid]['uuid'] = [
        '#type' => 'hidden',
        '#default_value' => $uuid,
      ];
      $element[$uuid]['remove'] = [
        '#type' => 'submit',
        '#value' => t('Remove'),
        '#submit' => [self::class . '::removePlugin'],
        '#ajax' => [
          'callback' => [self::class, 'removePluginAjax'],
          'event' => 'click',
          'wrapper' => $wrapper,
        ],
        '#name' => 'remove-plugin-' . $uuid,
        '#limit_validation_errors' => [array_slice($element['#parents'], 0, -1)],
      ];
    }

    // The addCleanValueKey method actually supports arrays (#3321409).
    $form_state->addCleanValueKey(array_merge($element['#array_parents'], ['new_plugin']));
    $form_state->addCleanValueKey(array_merge($element['#array_parents'], ['add']));
    $form_state->addCleanValueKey(array_merge($element['#array_parents'], ['plugin_group']));

    $element['new_plugin'] = [
      '#type' => 'plugin_select',
      '#title' => t('New plugin'),
      '#plugin_manager' => $element['#plugin_manager'],
      '#configuration' => $element['#configuration_add_form'] ?? FALSE,
      '#uuid' => $element['#uuid'] ?? 'no',
    ];
    $disabled = FALSE;
    if (!empty($element['#cardinality']) && $element['#cardinality'] != -1) {
      $disabled = count($element['#default_value']) >= $element['#cardinality'];
    }
    $element['add'] = [
      '#type' => 'submit',
      '#value' => isset($element['#plugin_name']) ? t('Add @n', ['@n' => $element['#plugin_name']]) : t('Add plugin'),
      '#submit' => [[self::class, 'addPlugin']],
      '#ajax' => [
        'callback' => [self::class, 'addPluginAjax'],
        'event' => 'click',
        'wrapper' => $wrapper,
      ],
      '#name' => 'add-plugin-' . ($element['#uuid'] ?? 'no'),
      '#disabled' => $disabled,
      '#limit_validation_errors' => [array_slice($element['#parents'], 0, -1)],
    ];


    return $element;
  }

  public static function addPlugin(&$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    $parents = array_slice($triggering_element['#parents'], 0, -1);
    $plugins = $form_state->getValue($parents);
    $cardinality = !empty(NestedArray::getValue($form, $parents)['#cardinality']) ? NestedArray::getValue($form, $parents)['#cardinality'] : -1;
    // The values are not cleaned yet - so we have to substract the numer of fields to clean 3.
    if ($cardinality > count($plugins) - 3 || $cardinality == -1) {
      $plugin = NestedArray::getValue($form_state->getUserInput(), array_merge($parents, ['new_plugin']));
      $element = [
        '#default_value' => $plugin,
        '#plugin_manager' => NestedArray::getValue($form, $parents)['#plugin_manager'],
      ];
      $plugin = PluginSelect::valueCallback($element, $plugin, $form_state);
      $uuid = \Drupal::service('uuid')->generate();
      $plugins[$uuid] = $plugin;
    }
    $form_state->setValue($parents, $plugins);
    $form_state->setRebuild();
  }

  public static function addPluginAjax(array &$form, FormStateInterface $form_state) {
    $button = $form_state->getTriggeringElement();
    $parents = array_slice($button['#parents'], 0 , -1);
    $element = NestedArray::getValue($form, $parents) ?? [];
    $cardinality = $element['#cardinality'] ?? -1 ;
    // The values are not cleaned yet - so we have to substract the numer of fields to clean 4.
    if ($cardinality <= count($form_state->getValue($parents)) - 4 && $cardinality != -1) {
      $element['add']['#attributes']['disabled'] = 'disabled';
    }
    return $element;
  }

  public static function removePlugin(&$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    $parents = array_slice($triggering_element['#parents'], 0, -2);
    $uuids = array_slice($triggering_element['#parents'], -2, -1);
    $plugins = $form_state->getValue($parents);
    unset($plugins[reset($uuids)]);
    $form_state->setValue($parents, $plugins);

    $form_state->setRebuild();
  }

  public static function removePluginAjax(array &$form, FormStateInterface $form_state) {
    $button = $form_state->getTriggeringElement();
    $parents = array_slice($button['#parents'], 0 , -2);
    $element = NestedArray::getValue($form, $parents) ?? [];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input) {
      foreach ($input as $uuid => $plugin_configuraion) {
        if (!is_array($plugin_configuraion) || empty($plugin_configuraion['id'])) {
          continue;
        }
        $p_element = [
          '#default_value' => $plugin_configuraion,
          '#plugin_manager' => $element['#plugin_manager'],
        ];
        // Use value callbacks for each plugin:
        $config = SinglePluginConfig::valueCallback($p_element, $plugin_configuraion, $form_state);
        $input[$uuid] = self::sortConfig($config);
      }

      $form_state->cleanValues();
    }

    return $input ?? [];
  }

  /**
   * Sort the config array.
   *
   * id and uuid shall be the first elements.
   *
   * @param array $config
   *   The plugin config
   * @return array
   *   The sorted config
   */
  public static function sortConfig(array $config): array {
    if (isset($config['uuid'])) {
      $uuid = $config['uuid'];
      unset($config['uuid']);
      $config = array_merge(['uuid' => $uuid], $config);
    }
    if (isset($config['id'])) {
      $id = $config['id'];
      unset($config['id']);
      $config = array_merge(['id' => $id], $config);
    }

    return $config;
  }

}
