<?php

namespace Drupal\plugin_config_form_builder\Element;

use Drupal\Component\Plugin\PluginInspectionInterface;
use Drupal\Component\Plugin\PluginManagerInterface;
use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormState;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Form\SubformState;
use Drupal\Core\Plugin\ContextAwarePluginInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\Core\Plugin\PluginWithFormsInterface;
use Drupal\Core\Render\Attribute\FormElement;
use Drupal\Core\Render\Element\FormElementBase;

/**
 * Provides a single_plugin_config element.
 */
#[FormElement('single_plugin_config')]
class SinglePluginConfig extends FormElementBase {

  /**
   * {@inheritdoc}
   */
  public function getInfo() {
    $class = static::class;
    return [
      '#input' => TRUE,
      '#process' => [
        [$class, 'processPlugin'],
      ],
      '#element_validate' => [
        [$class, 'validatePluginConfig'],
      ],
      '#theme_wrappers' => ['container'],
    ];
  }

  /**
   * Form element validation handler for #type 'plugin_config'.
   */
  public static function validatePluginConfig(&$element, FormStateInterface $form_state, &$complete_form) {
    $plugin_instance = static::getPlugin($element, $form_state, $complete_form);

    if ($plugin_instance instanceof PluginFormInterface) {
      $subform_state = SubformState::createForSubform($element, $element, $form_state);
      $form = [];
      static::getPluginForm($plugin_instance)->validateConfigurationForm($form, $subform_state);
    }
  }

  /**
   * Expands the element.
   */
  public static function processPlugin(&$element, FormStateInterface $form_state, &$complete_form) {
    /** @var \Drupal\Core\Plugin\Context\ContextRepositoryInterface $context_repository */
    $context_repository = \Drupal::service('context.repository');

    $element += [
      '#type' => 'fieldset',
      '#collapsible' => FALSE,
      '#prefix' => '<div id="' . $element['#id'] . '-wrapper">',
      '#suffix' => '</div>',
    ];

    $element['#tree'] = TRUE;
    $plugin = static::getPlugin($element, $form_state, $complete_form);
    // Add the plugin fields to the form.
    $subform_state = new FormState();
    $subform_state->setTemporaryValue('gathered_contexts', $context_repository->getAvailableContexts());
    $plugin_form = static::getPluginForm($plugin)->buildConfigurationForm([], $subform_state);
    $element += $plugin_form;

    return $element;
  }

  /**
   * Loads plugins from config.
   */
  public static function getPlugin(&$element, FormStateInterface $form_state, &$complete_form = []) {
    $plugin_configuration = $element['#default_value'];
    $plugin_configuration += NestedArray::getValue($form_state->getValues(), $element['#array_parents']) ?: [];

    /** @var PluginManagerInterface $plugin_manager */
    $plugin_manager = \Drupal::service($element['#plugin_manager']);

    return $plugin_manager->createInstance($element['#default_value']['id'], $plugin_configuration);
  }

  /**
   * Retrieves the plugin form for a given plugin.
   *
   * @param PluginInspectionInterface $plugin
   *   The plugin plugin.
   * @return \Drupal\Core\Plugin\PluginFormInterface
   *   The plugin form for the plugin.
   * @throws \Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException
   */
  public static function getPluginForm(PluginInspectionInterface $plugin) {
    if ($plugin instanceof PluginWithFormsInterface) {
      return \Drupal::service('plugin_form.factory')->createInstance($plugin, 'configure');
    }

    return $plugin;
  }

  /**
   * {@inheritdoc}
   */
  public static function valueCallback(&$element, $input, FormStateInterface $form_state) {
    if ($input) {
      /** @var PluginManagerInterface $plugin_manager */
      $plugin_manager = \Drupal::service($element['#plugin_manager']);
      if (isset($element['#default_value']['id']) && $plugin_manager->hasDefinition($element['#default_value']['id'])) {
        // The user input might be empty.
        if (empty($input)) {
          $input = [];
        }
        $plugin = $plugin_manager->createInstance($element['#default_value']['id'], $input + $element['#default_value']);

        // Call the plugin submit handler.
        $parent_form = [$element];
        $parent_form['#parents'] = [];
        $subform_state = new FormState();

        $input += $plugin->defaultConfiguration();
        $subform_state->setValues($input);
        static::getPluginForm($plugin)->submitConfigurationForm($element, $subform_state);

        // If this block is context-aware, set the context mapping.
        if ($plugin instanceof ContextAwarePluginInterface) {
          $plugin->setContextMapping($subform_state->getValue('context_mapping', []));
        }

        // Replace the configuration with the configuration from the plugin.
        $input = $plugin->getConfiguration();
      }
    }

    return $input ?? [];
  }

}
